! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

! Module      : ommag_mod
!
! Authors     : Tim Harries (MSSL)
!             : Phil Williams (MSSL)
!
! Description : Convert from count-rates to magnitudes in the instrumental system.
!
! Comments    :
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
!
! History : Created  : March 98 - TJH
!           Modified : 20 May 99 - Spring clean of code.
!           Modified : 22/8/2000 (CB) - added call to CAL_omGetColorTrans
!                      to convert instrumental magnitudes to standard magnitudes.
!           modified : 19/9/2000 (AMK) - corrected call to CAL_omPhotoMag. Parameters 
!                      were set incorrectly.
!    4th. October 2000 - AMK
!    Replace the call getFIlterID in mssllib by the OAL call filterIdToInt
!    16th May 2001 (CB) : To comply with the CAL definitions of magnitudes, etc,
!                         the counts (after having been corrected for coincidence
!                         losses, etc, have to be scaled from the nominal 12 pixel
!                         aperture radius to the CAL specified radius for the
!                         particular filter (12 for U,B and V, 35 for UV filters).
!                         Hence no correction for optical filters.
!   7th June 2001 (CB) :  Modified to read in the source-list column of source
!                         aperture-radii produced by omdetect. If a particular
!                         radius is less than the 12-pixel value required for the
!                         correction of coincidence losses, etc, then the cal routine
!                         "CAL_omPsfEncircledEnergy" is used to extrapolate the counts.
!                         A further extrapolation is done for UV filters- from 12 pixels
!                         to 35 pixels.  
!
!
!  6th of july 2001 VNY   small correction has been made (PI) 
!  5tn Decemmer 2001 (CB) BINBPE now read in for calculating the frame-time
!  18thJanuary 2002 (CB) 0.02 magnitude offset added to take account of systematic errors
!  14 th May 2002 (CB) Extra columns added to store the scorrected count rates and some
!                      key words added to source-list file 
!  16th August 2002 (CB) Since omdetect has been modified to do the coincidence-loss
!   correction ommag has been modified not to do this one.
!  2013-10-15 (vny) modifying to get more rigourous corrected count rate error calculation
!  2013-10-29 since the column CORR_RATE is read to the array rawRate, the name of 
! this array is changed to initialCorrRate
! 2013-10-31 fixed a bug in the minimum-searching subroutine "brent", which was
! leading to a completely wrong calculation of the coincidence-loss error-bars
! 2013-11-25 Changing the order of calculation of the error bars: calculating 
! time-dependent scaling AFTER the Simon's errors
! 2013-11-26: modified the column CORR_BKG_RATE_ERR by applying the time-dependent
! correction factor; fine-tuned the interpolation of binomial error bars
!********************************************************************************
module ommag_mod

   use param
   use cal
   use dal
   use errorHandling
   use frameTime_mod
   use omallocation_mod
   use omutility_mod
   use ommagnitude_mod
   use omcoinccorr_mod
   use countrateerror_mod
   use omconstants_mod
   implicit none

   private

   public ommag

contains

!---------------------------------------------------------------------
subroutine ommag
!---------------------------------------------------------------------
    character(255) :: set                                         ! OM OSW Source List filename.
    character(8) :: dataMode                                      ! Input file data mode.
    character(9) :: filter                                        ! Keywords.

    type(DataSetT) :: inDataSetH                                  ! Input dataset handle.
    type(TableT)   :: inTableH                                    ! Input table handle.
    !integer, parameter :: noColumns = 7
    integer, parameter :: noColumns = 18
    type(ColumnT), dimension(noColumns) :: inColumnH                          ! Input image column handle.

    integer(int32) :: nSources                                   ! Number of sources.
    integer(int32) :: ccdMode                                    ! OM CCD data mode.
   
    integer(int32) :: i                                          ! Count.
    integer(int32) :: filterIdInt
    integer(int8), dimension(:), pointer :: extended

    real(double) :: corrRate                                      ! Corrected count rate.
    real(double) :: corrRateErrPlus                               ! Corrected rate error + (asymmetric).
    real(double) :: corrRateErrMinus                              ! Corrected rate error - (asymmetric).
    real(double) :: corrMagErrPlus                                ! Corrected magnitude error + (asymmetric).
    real(double) :: corrMagErrMinus                               ! Corrected magnitude error - (asymmetric).
    real(single) :: countFrameRatio = 0.5
    real(single) :: frameRate 
    
    real(single), dimension(:), pointer :: rawRate => null()              ! Raw rates pointer.
    real(single), dimension(:), pointer :: rawRateErr => null()           ! Raw rates error pointer.
    real(single), dimension(:), pointer :: initialCorrRate => null()              ! Raw rates pointer.
    real(single), dimension(:), pointer :: initialCorrRateErr => null()           ! Raw rates error pointer.

    real(single), dimension(:), pointer :: correctedRate => null()        ! Corrected rates pointer.
    real(single), dimension(:), pointer :: correctedRateErr => null()     ! Corrected rates error pointer.

    !real(single), dimension(:), pointer :: correctedBkgRate => null()        ! Corrected rates pointer.
    !real(single), dimension(:), pointer :: correctedBkgRateErr => null()     ! Corrected rates error pointer.

    real(single), dimension(:), pointer :: sourceXCoord => null()         ! Source x-coordinate.
    real(single), dimension(:), pointer :: sourceYCoord => null()         ! Source y-coordinate. 
    
 
    real(single), dimension(:), pointer :: significance => null()         ! Source significance.


    real(single), dimension(:), pointer :: rawBackGroundRate => null()       ! Source background raw rate pointer.
    real(single), dimension(:), pointer :: rawBackGroundRateErr => null()       ! Source background raw rate pointer.
    !real(single), dimension(:), pointer :: correctedBackGroundRate => null() ! Source background corrected rate pointer.
    real(single), dimension(:), pointer :: correctedBkgRate => null() ! Source background corrected rate pointer.
    real(single), dimension(:), pointer :: correctedBkgRateErr => null()

    real(single), dimension(:), pointer :: initialCorrectedBkgRate => null()
    real(single), dimension(:), pointer :: initialCorrectedBkgRateErr => null()

    real(single), dimension(:), pointer :: psf2Correction => null()         ! psf correction
 
    real(single), dimension(:), pointer :: mag => null()                  ! Magnitude pointer.
    real(single), dimension(:), pointer :: magErr => null()               ! Magnitude error pointer.
    !*******************************************************************************
    ! 2013-10-05 adding an extra column for checking the count-rate error calculation
    real(single), dimension(:), pointer :: corrRateErrA => null() ! modified coorected rate error (Antonio)
    !
    real(single), dimension(:), pointer :: rateA => null()  ! raw src rate
    real(single), dimension(:), pointer :: rateAerr => null() ! raw src rate error
    real(single), dimension(:), pointer :: bkgRateA => null() ! raw bkg rate
    real(single), dimension(:), pointer :: bkgRateAerr => null() ! raw bkg rate error

    real(single), dimension(:), pointer :: psf1corrA => null() 
    real(single), dimension(:), pointer :: spbCoilossA => null()
    real(single), dimension(:), pointer :: bkCoilossA => null()
    real(single), dimension(:), pointer :: apertureA => null() 

    real(single), dimension(:), pointer :: majorA => null() 
    real(single), dimension(:), pointer :: minorA => null() 

    !*******************************************************************************
    real(double) :: frameTime                                    ! OM CCD frame time.
    real(double) :: deadFrac = 0.0                               ! Dead time in units of CCD frame time.
   
    real(double)              :: detectionLimitCountRate         ! The detection limit count rate (counts/second)
    real(double)              :: detectionLimitMagnitude         ! The detection limit magnitude
    real(single) :: error = 0.0  
    real(single) :: calApertureRadius = 12.0          ! Radius used for coincidence losses, etc   
    real(single) :: photometryApertureRadius          ! Radius used for converting counts to magnitudes
    logical :: scaleCounts = .false.
    real(double) :: countsRatio = 1.0_double
    real(double) :: scaleCountsRatio = 1.0_double
    real(single) :: exposure   
    integer           :: position
    real(single) :: nullValue = -999.999  
    real(double) :: deadFractionCorrection
  
    real(single) :: estimatedRateErr

    real(double) :: mjd
    real(single) :: tdpCorrFactor
    !real(single) :: deadTimeCorrection

    !real(single) :: psf2factor, psf1factor
    !real(single) :: psf2corr
    !real(single) :: apertureRatio
    !real(single) :: aux, auxRadius
    real(single) :: exposureM

    !************************
    ! Read in parameters.
    !************************
    set = stringParameter("set")

    
    ! Read in the OM OSW Source List file.
    inDataSetH = dataSet(set, modify)
   
    !write(*,*) "L173 ============ starting ommag ==================="
    if (.not. hasBlock(inDataSetH, "SRCLIST")) then
        !************************************
        ! Incompatible file - can't continue
        !************************************
        write (errstr, *) "Could not find any compatibile blocks in set ", set
        call fatal("badFileType")
    end if

    !**************************************************************
    ! Read in the detection limit count rate from the PRIMARY header
    !**************************************************************
    detectionLimitCountRate = real(real32Attribute(inDataSetH, "CRLIMIT"))
           
    ! Get a handle on the source list binary table extension.
    inTableH = table(inDataSetH, "SRCLIST")


    !******************************************************************
    ! Get the number of rows in the source list binary table extension.
    !******************************************************************
    nSources = numberOfRows(inTableH)
    
    write (errstr, *) "The number of sources :  ", nSources
    call message(AppMsg, Verbose)

    if (nSources .le. 0) then
        call release(inTableH)
        call release(inDataSetH)
        call warning("No sources in source-list file")
        stop
    end if        
 

    ! Read in the source list table extension keywords.
    dataMode = stringAttribute(inDataSetH, "DATAMODE")
    filter = stringAttribute(inDataSetH, "FILTER")

    exposure = real32Attribute(inDataSetH, "EXPOSURE")
    
    frameTime = real64Attribute(inDataSetH, "FRAMTIME") / 1000.0
    deadFrac = real64Attribute(inDataSetH, "DEADFRAC")

    !*************************************************************************
    ! Need the modified Julian date for time-dependent sensitivity correction
    !************************************************************************
     mjd = ModifiedJulianDate(trim(stringattribute(inDataSetH, "DATE-OBS")))

    !write(*,*) "L215 dataMode=", dataMode, " filter=", filter, " exposure=", exposure
    !write(*,*) "L216 frameTime=", frameTime, " deadFrac=", deadFrac
    write (errstr, *) "Observation date : ", trim(stringattribute(inDataSetH, "DATE-OBS")), &
         " MJD=", mjd
    call message(AppMsg, Verbose)
    write (errstr, *) "OM filter: ", filter, " exposure : ", exposure
    call message(AppMsg, Verbose)

    !********************************************************************
    ! 2013-10-15, 2013-10-29 modifying the source corr.error calcilation
    !********************************************************************
    deadFractionCorrection=1.0/(1.0-deadFrac)
    
    !write(*, *) "L229 Date_OBS=", trim(stringattribute(inDataSetH, "DATE-OBS")), " filter =", trim(filter), &
    !     " mjd=", mjd
  
    deadFractionCorrection = 1.0_double / (1.0_double - deadFrac)
    exposureM=exposure
    if (deadFractionCorrection .gt. 0.0) then
       exposureM=exposureM/deadFractionCorrection
    endif
    
    frameRate = 1.0 / frameTime

    !********************
    ! Get the ccd mode.
    !********************
    if (trim(dataMode) .eq. "IMAGING") then
        ccdMode = IMAGE
    else
        ccdMode = FAST
    end if

    !**************************************
    ! Get the filter Id integer value
    !**************************************
    filterIdInt = filterIdToInt(filter)

    !*******************
    ! Set the CAL state
    !********************
    call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)

    !*********************************************************
    ! Time-dependent sensitivity correction-factor
    !*********************************************************
    tdpCorrFactor = TDPCorrectionFactor(mjd, filter)

    !write(*,*) "L265 tdpCorrFactor=", tdpCorrFactor
    write (errstr, *) "Time-dependent correction factor :  ", tdpCorrFactor
    call message(AppMsg, Verbose)

    !*************************************************************
    ! 17th May 2001 - Due to the changes to the photometry in
    ! omdetect, now need to scale the source counts from the 12
    ! pixel aperture radius to that specified by the CAL.
    ! For U, B and V there will be no correction.
    !************************************************************

    photometryApertureRadius = CAL_omGetApertureRadius()
    !write(*,*) "L273 photometryApertureRadius=", photometryApertureRadius, " calApertureRAdius=", &   
    !     calApertureRadius

    if (photometryApertureRadius .gt. calApertureRadius) then
        !**********************************************
        ! Remember that counts have to be extrapolated.
        !**********************************************
        scaleCounts = .true.
        scaleCountsRatio = (photometryApertureRadius / calApertureRadius) ** 2 
    end if
    !write(*,*) "L283 scaleCountsRatio=", scaleCountsRatio

    !******************************************************
    ! Calculate the instrumental limiting magnitude from
    ! the detection limit count rate
    !******************************************************
    if (detectionLimitCountRate .gt. 0.0) then      
        if (frameTime .gt. 0.0) then 
           !write(*,*) "L291 applying the routine CorrectForCoincidenceLoss"
            corrRate = CorrectForCoincidenceLoss(detectionLimitCountRate, error, frameTime, &
                  corrRateErrMinus,  corrRateErrPlus)
            !write(*,*) "L294 detectionLimitCountRate=", detectionLimitCountRate, &
            !     " corrRate=", corrRate
        else
            corrRate = detectionLimitCountRate
        end if
        !write(*,*)  "L299 (before applying the dead-time and time-dependent corr.) corrRate=", corrRate

        !************************************************
        ! Correct for dead-time
        !************************************************
        corrRate = corrRate * deadFractionCorrection 
        corrRateErrMinus = corrRateErrMinus * deadFractionCorrection 
        corrRateErrPlus = corrRateErrPlus * deadFractionCorrection 

        !*********************************************************
        ! Time-dependent sensitivity correction
        !******************************************************** 
        corrRate = corrRate * tdpCorrFactor 
        corrRateErrMinus = corrRateErrMinus * tdpCorrFactor 
        corrRateErrPlus = corrRateErrPlus * tdpCorrFactor 
     
        if (corrRate .gt. 0.0) then
            !**********************************************************
            ! May need to scale the counts from 12 to 35 pixels for uv
            !*********************************************************
            if (scaleCounts) then
                countsRatio = PSFCountsRatio(photometryApertureRadius, calApertureRadius, &
                                     1024.5, 1024.5, countFrameRatio)
                corrRate = corrRate * countsRatio
                corrRateErrMinus = corrRateErrMinus * countsRatio
                corrRateErrPlus = corrRateErrPlus * countsRatio
  
            end if
             
            call CAL_omPhotoMagnitude(corrRate, detectionLimitMagnitude, corrRateErrMinus, &
                                   corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
        else
            detectionLimitMagnitude = 0.0d0
        end if
    else
        detectionLimitmagnitude = 0.0
    end if

    write (errstr, *) "Detection limit magnitude :  ", detectionLimitmagnitude
    call message(AppMsg, Verbose)
    !write(*,*) "L337 detectionLimitMagnitude=", detectionLimitMagnitude
    call setAttribute(inDataSetH, "MAGLIMIT", real(detectionLimitMagnitude), &
             "mag", "detection limiting magnitude (instrumental)")

    !**************************************************
    !
    ! Get the data from the source list file
    ! The rates have been corrected for significance-losses
    ! by omdetect- referred to as RAW_RATES
    !
    !**************************************************
    write(*,*) "L348 getting the column data"

    initialCorrRate => GetRealColumn(inTableH, "CORR_RATE", inColumnH(1))
    initialCorrRateErr => GetRealColumn(inTableH, "CORR_RATE_ERR", inColumnH(2))

    mag => AllocateSingle(0, nSources - 1, "mag")
    magErr => AllocateSingle(0,  nSources - 1, "magErr")

    initialCorrectedBkgRate => GetRealColumn(inTableH, "CORR_BACKGROUND_RATE", inColumnH(3))
    initialCorrectedBkgRateErr => GetRealColumn(inTableH, "CORR_BKG_RATE_ERR", inColumnH(4))

    sourceXCoord => GetRealColumn(inTableH, "XPOS", inColumnH(5))
    sourceYCoord => GetRealColumn(inTableH, "YPOS", inColumnH(6))

    inColumnH(7) = column(inTableH, "EFLAG", modify)
    extended => int8Data(inColumnH(7))

    significance => GetRealColumn(inTableH, "SIGNIFICANCE", inColumnH(8))

    !tttttttttttttttttttttttttttttt
    ! 2013-10-15 extra columns needed for the calculation of the modified count rate error
    !rateA => GetRealColumn(inTableH, "RATE", inColumnH(9))
    !rateAerr => GetRealColumn(inTableH, "RATE_ERR", inColumnH(10))
    !bkgRateA => GetRealColumn(inTableH, "BACKGROUND_RATE", inColumnH(11))
    !bkgRateAerr => GetRealColumn(inTableH, "BKG_RATE_ERR", inColumnH(12))

    rawRate => GetRealColumn(inTableH, "RATE", inColumnH(9))
    rawRateErr => GetRealColumn(inTableH, "RATE_ERR", inColumnH(10))
    rawBackGroundRate => GetRealColumn(inTableH, "BACKGROUND_RATE", inColumnH(11))
    rawBackGroundRateErr => GetRealColumn(inTableH, "BKG_RATE_ERR", inColumnH(12))
    !tttttttttttttttttttttttttttttttttttt

    psf1corrA => GetRealColumn(inTableH, "PSF1_CORR", inColumnH(13))
    spbCoilossA => GetRealColumn(inTableH, "SPB_COILOSS", inColumnH(14))
    bkCoilossA => GetRealColumn(inTableH, "BK_COILOSS", inColumnH(15))
    apertureA => GetRealColumn(inTableH, "APERTURE", inColumnH(16))

    majorA => GetRealColumn(inTableH, "FWHM_MAJ", inColumnH(17))
    minorA => GetRealColumn(inTableH, "FWHM_MIN", inColumnH(18))

    !******************************************************************
    ! Loop through all the sources- 
    ! Correct each count-rate for the effect of detector 
    ! lifetime and ageing, then calculate the magnitude.
    !******************************************************************
         
    !****************************************************
    !
    ! Arrays for storing the corrected count rates, etc
    !
    !**************************************************** 
    correctedRate => AllocateSingle(0, nSources - 1, "correctedRate")  
    correctedRate = nullValue
    correctedRateErr => AllocateSingle(0, nSources - 1, "correctedRateErr")  
    correctedRateErr = nullValue

    !********************************************
    ! 2013-10-15 vny: adding an extra corr-rate-err
    !**********************************************
    corrRateErrA => AllocateSingle(0, nSources - 1, "corrRateErrA")  
    corrRateErrA = nullValue
 
    correctedBkgRate => AllocateSingle(0, nSources - 1, "correctedBkgRate")  
    correctedBkgRate = nullValue
      
    ! 2013-11-26 adding the corrected background error
    correctedBkgRateErr => AllocateSingle(0, nSources - 1, "correctedBkgRateErr")  
    correctedBkgRateErr = nullValue

    psf2Correction => AllocateSingle(0, nSources - 1, "psf2Correction")  
    psf2Correction = nullValue

    !********************************************************
    ! Loop through the sources, do the dead-time correction
    ! and then compute the instrumental magnitude
    !********************************************************

    !tttttttttttttttttttttttttttttttttttttt
    correctedRate = initialCorrRate 
    correctedRateErr = initialCorrRateErr
 
    correctedBkgRate = initialCorrectedBkgRate
    correctedBkgRateErr = initialCorrectedBkgRateErr

    !correctedRate = rawRate 
    !correctedRateErr = rawRateErr
 
    !correctedBkgRate = rawBackGroundRate
    !correctedBkgRateErr = rawBackGroundRateErr
    !ttttttttttttttttttttttttttttttttttttttttttttt


    mag = nullValue
    magErr = nullValue


    !write(*,*) "*******************************************************************************"
    !write(*,*) "L470 deadFractionCorrection=", deadFractionCorrection, " tdpCorrFactor=", tdpCorrFactor

    write (errstr, *) "Dead-fraction correction :  ", deadFractionCorrection
    call message(AppMsg, Verbose)

    !write(*,*) "L449 frameTime=", frameTime, " photometryApertureRadius=", photometryApertureRadius, &
    !     " calApertureradius=", calApertureradius
    !write(*,*) "L451 countFrameRatio=", countFrameRatio
 !   write(*, *) "*****************************************************************************"
 !   write(*, *) " rtA bkgRtA psf1crA spbClA bkClA apA cRErA cntRat psf2cr psf2f psf1f aprat"
    !write(*, *) "*****************************************************************************"
    !tttttttttttt
    !write(*, *) "*****************************************************************************"
    !write(*, *) " x   y   initCorrRate  initCorrRateErr  correctedRate correctedRateErr"
    !write(*, *) "*****************************************************************************"
    !do i = 0, nSources - 1
       !write(*,*) ".................... i=",i
       !write(*, *) i,  sourceXCoord(i),  sourceYCoord(i),  initialCorrRate(i),  initialCorrRateErr(i), &
       !     correctedRate(i), correctedrateErr(i) 
!       !***********************************************
!       ! 2013-10-15 vny: modified corrected rate error
!       !***********************************************
!       countsRatio = PSFCountsRatio(photometryApertureRadius, calApertureRadius, &
!            sourceXCoord(i), sourceYCoord(i), countFrameRatio)
!       write(*,*) "L461 x=", sourceXCoord(i), " y=", sourceYCoord(i), " ap=", apertureA(i), &
!            " countsRatio=", countsRatio
!
!       psf2corr=(countsRatio - 1.0) * 100.0
!       psf2factor= countsRatio
!       psf1factor= 1.0+psf1corrA(i)/100.0
!       apertureRatio=1.0
!       if (apertureA(i) .gt. 0.0) then
!         apertureRatio=calApertureRadius/apertureA(i)
!          apertureRatio=apertureRatio*apertureRatio
!       else
!          !apertureRatio=-1.0
!          !
!          !auxRadius=sqrt(2.0)*(majorA(i)+minorA(i))
!          auxRadius=3.0*(majorA(i)+minorA(i))/2.354 ! 3-sigma aperture radius
!          if (auxRadius .gt. 0.0) then
!             apertureRatio=calApertureRadius/auxRadius
!          else
!             apertureRatio=calApertureRadius/10.0
!          endif
!          apertureRatio=apertureRatio*apertureRatio
!       endif
!
!       write(*,*) "L484 apertureRatio=", apertureRatio
!
!       if (apertureRatio .lt. 0.0) then
!          corrRateErrA(i)=correctedRateErr(i) ! this is for extended sources
!       else
!          aux=spbCoilossA(i)*spbCoilossA(i)*(rateA(i)*psf1factor*psf2factor + &
!               bkgRateA(i)*apertureRatio) + &
!               bkCoilossA(i)*bkCoilossA(i)*bkgRateA(i)*apertureRatio
!          aux=sqrt(aux)
!          if (exposureM .gt. 0.0) then
!             aux=aux/sqrt(exposureM)
!          endif
!          corrRateErrA(i)=correctedRateErr(i)          
!          !correctedRateErr(i)=aux*psf2factor*deadFractionCorrection*tdpCorrFactor
!
!       endif
!
!       write(*, *) i,  rateA(i),  bkgRateA(i),  psf1corrA(i), spbCoilossA(i), bkCoilossA(i), apertureA(i), &
!            corrRateErrA(i),  countsRatio, psf2corr, psf2factor, psf1factor, apertureRatio
    !end do
    !write(*, *) "*****************************************************************************"
    !tttttttttttt

    !**********************************************************
    ! Compute Simon's errors
    !********************************************************* 

    !write(*, *) "L511 ommag frametime=", frameTime
    !***************************************************
    ! 2013-10-16 vny: excluding Simon errors as this gives a weird mag-magerr dependence
    !*************************************************
!ttttttttttttttttttttttttttttttt
    if (frameTime > 0.0) then
       !write(*,*) "L516 calling the routine ComputeSimonErrors"       
       !call ComputeSimonErrors(nSources, correctedRate, correctedRateErr, &
       !     correctedBkgRate, exposure, deadfrac, frameTime)
       !*using the formula directly
       call ComputeSimonErrors2(nSources, rawRate, rawRateErr, rawBackGroundRate, &
            correctedRate, correctedRateErr, &
            correctedBkgRate, exposure, deadfrac, frameTime)
    end if
!tttttttttttttttttttttttttttttt    

    !*********************************************************
    ! Applying time-dependent sensitivity correction
    ! 2013-11-26 vny: applying this after the Simon's errors
    !******************************************************** 

    !write(*, *) "*****************************************************************************"
    !write(*, *) "L566 tdpcorrfactor=", tdpcorrfactor, " exposure=", exposure, " secs", &
    !     " deadfrac=", deadfrac, " frametime=", frametime
    !write(*, *) "*****************************************************************************"
    !write(*, *) " x   y   corrRate  corrRateErr corrBkgRate  corrBkgRateErr "
    !write(*, *) "*****************************************************************************"

    !---------------------
    ! 2013-12-02 limiting the corrected count rate error by the magnitude of the 
    ! corrected count rates
    !---------------------
    do i = 0, nSources - 1

       !write(*, *) i,  sourceXCoord(i),  sourceYCoord(i),  &
       !      correctedRate(i), correctedRateErr(i), correctedBkgRate(i), &
       !      correctedBkgRateErr(i)
 
       ! 2014-01-30 vny: estimating the source rate error as it should be based
       ! on the formula sqrt(s+2*b)
       if (.not. isNullValue(correctedBkgRate(i)) .and. correctedRate(i) .gt. 0.0) then
          estimatedRateErr=(correctedRate(i)+2.0*correctedBkgRate(i))*exposure
          estimatedRateErr=sqrt(estimatedRateErr)/exposure 
          !write(*,*) "L587 estimatedRateErr=", estimatedRateErr           
       endif

       if (correctedRateErr(i) .gt. correctedRate(i)) then
          !write(*,*) "L583 correctedRateErr is larger than correctedRate" 
          !---------------------
          ! 2014-01-30 limiting the corrected count rate error by the magnitude of the 
          ! corrected count rates
          !---------------------
          !correctedRateErr(i)=nullValue!correctedRate(i)
          correctedRateErr(i)=correctedRate(i)
          !write(*,*) "L590 replaced correctedRateErr=", correctedRateErr(i) 
          if (.not. isNullValue(correctedBkgRate(i)) .and. correctedRate(i) .gt. 0.0) then
             correctedRateErr(i)=(correctedRate(i)+2.0*correctedBkgRate(i))*exposure
             correctedRateErr(i)=sqrt(correctedRateErr(i))/exposure 
             !write(*,*) "L594 recalculated correctedRateErr=", correctedRateErr(i)           
          endif
       endif
       
      if (correctedRateErr(i) .gt. 2.0*estimatedRateErr) then
         !write(*,*) "L607 correctedRateErr >2 of expected (fixing)"
         correctedRateErr(i)=estimatedRateErr
         !write(*,*) "L609 fixed correctedRateErr(i)=", correctedRateErr(i)
      endif

       if (correctedRate(i) > 0.0) then
          correctedRate(i) = correctedRate(i) * tdpCorrFactor
          
          if (.not. isNullValue(correctedRateErr(i))) then
             !write(*,*) "      L595 correctedRateErr=", correctedRateErr(i) 
             correctedRateErr(i) = correctedRateErr(i) * tdpCorrFactor
             !write(*,*) "      L597 correctedRateErr=", correctedRateErr(i) 
          endif
          if (.not. isNullValue(correctedBkgRate(i))) then
             correctedBkgRate(i) = correctedBkgRate(i) * tdpCorrFactor
          endif
          if (.not. isNullValue(correctedBkgRateErr(i))) then
             correctedBkgRateErr(i) = correctedBkgRateErr(i) * tdpCorrFactor
          endif

          !tttttttttttttttttt
          !correctedRate(i) = correctedRate(i) 
          !correctedRateErr(i) = correctedRateErr(i) 
          !correctedBkgRate(i) = correctedBkgRate(i) 
          !correctedBkgRateErr(i) = correctedBkgRateErr(i)
          !ttttttttttttttttt
  
        end if

        !write(*,*) "       L613 correctedRateErr=", correctedRateErr(i) 
    end do

     !ttttttttttttttt
     !do i = 0, nSources - 1
     !   corrRateErrA(i)=correctedRateErr(i)
     !enddo
     !ttttttttttttttt
    !write(*, *) "**************************************************************************"
    do i = 0, nSources - 1
       !write(*,*) "L615 i=", i, "-----------------"
       !write(*,*) "L616 correctedRate=", correctedRate(i), " correctedRateErr=", correctedRateErr(i) 
       !write(*,*) "L617 correctedBkgRate=", correctedBkgRate(i), " correctedBkgRateErr=", correctedBkgRateErr(i) 
       if (correctedRate(i) .gt. 0.0 .and. correctedRate(i) .le. 1000.0 ) then
          call ComputeMagnitude(correctedRate(i), correctedRateErr(i),             &
               !correctedBackgroundRate(i),  correctedBackgroundRateErr, &
               !**************************************************
               ! 2014-01-30 VNY: a bug: correctedBackgroundRateErr was not replaced with
               ! correctedBkgRateErr (fixing)
               !correctedBkgRate(i),  correctedBackgroundRateErr, &
               !************************************************************
               correctedBkgRate(i),  correctedBkgRateErr(i), &
               sourceXCoord(i), sourceYCoord(i),                        &
               scaleCounts, extended(i),                                &
               photometryApertureRadius, calApertureRadius,             &
               frameTime, deadFrac,                                     &
               mag(i), magErr(i), psf2Correction(i))
       else
          !if (.not. isNullValue(significance(i)) ) significance(i) = -significance(i) 
          !write(*,*) "L627 corrRAte(i)<0"
       end if
       

!rawBackgroundrate(i),  correctedBackgroundRate(i), mag(i), magerr(i), extended(i) 
!100 format(i6, 2f8.2, 6f12.2, f8.2, f9.3, i6)
    end do

   
    !*************************
    ! Release handles on data
    !***********************

    do i = 1, noColumns
        call release(inColumnH(i))
    end do

    !**************************************************
    ! Add the columns for the corrected count rates,
    ! after the significance column
    !**************************************************
    position = columnNumber(inTableH, "SIGNIFICANCE")
    !write(*,*) "L490 position=", position
    call AddReal32Column(inTableH, correctedRate, nSources, "CORR_RATE", &
            !"counts/sec",  "Corrected for coincidence losses, deadfraction, etc", nullValue, position)
         "counts/sec",  "Corrected for coincidence losses, deadfraction, etc", nullValue, position+1)

    call AddReal32Column(inTableH, correctedRateErr, nSources, "CORR_RATE_ERR", &
            !"counts/sec", "Corrected for coincidence losses, deadfraction, etc" , nullValue, position + 1)
         "counts/sec", "Corrected for coincidence losses, deadfraction, etc" , nullValue, position + 2)
 
    !****************************************************************
    ! Add extra column for corrRateErrA
    !**************************************************************** 
!    call AddReal32Column(inTableH, corrRateErrA, nSources, "CORR_ERRA", &
!            "counts/sec", "Corrected rate error modified)", nullValue, position + 3)
!
    !call AddReal32Column(inTableH, correctedBackgroundRate, nSources, "CORR_BACKGROUND_RATE", &
    call AddReal32Column(inTableH, correctedBkgRate, nSources, "CORR_BACKGROUND_RATE", &
            "counts/sec", "Corrected for coincidence losses, deadfraction, etc", nullValue, position + 3)
            !"counts/sec", "Corrected for coincidence losses, deadfraction, etc", nullValue, position + 5)
 
     call AddReal32Column(inTableH, correctedBkgRateErr, nSources, "CORR_BKG_RATE_ERR", &
            "counts/sec", "Corrected for coincidence losses, deadfraction, etc", nullValue, position + 4)

    call AddReal32Column(inTableH, psf2Correction, nSources, "PSF2_CORR", &
            "%", "PSF2 correction (12pixels to cal radius)", nullValue, position + 5)
            !"%", "PSF2 correction (12pixels to cal radius)", nullValue, position + 6)
            !"%", "PSF2 correction (12pixels to cal radius)", nullValue, position + 4)

    !******************************************************************************
    ! Time-dependent sensitivity correction factor
    !*****************************************************************************

    call addcomment(inTableH, "TDS_CORR is the correction factor applied to the corrected count-rates")   
    call addcomment(inTableH, "to correct for the OM sensitivity degradation")

 
    call setattribute(inTableH, "TDS_CORR", tdpCorrFactor, "Sensitivity degradation correction fator")
 


    !****************************************************************
    ! Add the magnitudes
    !****************************************************************
    if ( hasColumn(inTableH, "BK_COILOSS")) then 
        position = columnNumber(inTableH, "BK_COILOSS") + 1
    else
        position = columnNumber(inTableH, "CORR_RATE") + 1
    end if
    call AddReal32Column(inTableH, mag, nSources, "MAG", &
            "mag", "Instrumental magnitude", nullValue, position)
    deallocate(mag)

    !****************************************************************
    ! Add the magnitude errors
    !**************************************************************** 
    call AddReal32Column(inTableH, magErr, nSources, "MAGERR", &
            "mag", "Instrumental magnitude error", nullValue, position + 1)
 
    deallocate(corrRateErrA)

    deallocate(magErr)

    call release(inTableH)


    call release(inDataSetH)

    deallocate(correctedRate)
    deallocate(correctedRateErr)
    !deallocate(correctedBackgroundRate)
    deallocate(correctedBkgRate)

    deallocate(psf2Correction)
   
end subroutine ommag

!****************************************
! Function GetRealColumn
! Returns an array containing the data
! in the named column of the given table
!***************************************
Function GetRealColumn(inTableH, name, columnH)  result(dataPtr)  
    type(TableT), intent(in) :: inTableH
    character(*), intent(in) :: name
    type(ColumnT), intent(inout) :: columnH
    real(single), dimension(:), pointer :: dataPtr                ! Magnitude error pointer.
   
    columnH = column(inTableH, name, modify)
    dataPtr => real32Data(columnH)

end function GetRealColumn


end module ommag_mod











